# -*- coding: utf-8 -*-
# @File    :   __init__.py
# @Time    :   2025/01/22 21:11:47
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
taken from https://github.com/THU-KEG/KoPL 

修改了 data.py 中 concept 部分的处理
'''