# -*- coding: utf-8 -*-
# @File    :   common.py
# @Time    :   2023/05/22 16:09:01
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    通用的数据集分词处理函数
'''
from ..register import registers

@registers.process_function.register("test_function")
def test_function(data, *args, **kwargs):
    return data 